# -*- coding: utf-8 -*-
# https://en.wikipedia.org/wiki/Unicode_block

from typing import Final, NamedTuple, Sequence, Tuple

from cvp.strings.unicode.planes import (
    BMP,
    PUA_A,
    PUA_B,
    SIP,
    SMP,
    SSP,
    TIP,
    UnicodePlane,
)


class UnicodeBlock(NamedTuple):
    plane: UnicodePlane
    name: str
    begin: int
    end: int

    @property
    def range(self) -> Tuple[int, int]:
        return self.begin, self.end

    def __repr__(self):
        return f"{self.plane}: U+{self.begin:04X} - U+{self.end:04X}"


def _block(
    plane: UnicodePlane,
    begin: int,
    end: int,
    block_name: str,
) -> UnicodeBlock:
    return UnicodeBlock(plane, block_name, begin, end)


# noinspection SpellCheckingInspection
BMP_BLOCKS: Final[Sequence[UnicodeBlock]] = (
    _block(BMP, 0x0000, 0x007F, "Basic Latin"),
    _block(BMP, 0x0080, 0x00FF, "Latin-1 Supplement"),
    _block(BMP, 0x0100, 0x017F, "Latin Extended-A"),
    _block(BMP, 0x0180, 0x024F, "Latin Extended-B"),
    _block(BMP, 0x0250, 0x02AF, "IPA Extensions"),
    _block(BMP, 0x02B0, 0x02FF, "Spacing Modifier Letters"),
    _block(BMP, 0x0300, 0x036F, "Combining Diacritical Marks"),
    _block(BMP, 0x0370, 0x03FF, "Greek and Coptic"),
    _block(BMP, 0x0400, 0x04FF, "Cyrillic"),
    _block(BMP, 0x0500, 0x052F, "Cyrillic Supplement"),
    _block(BMP, 0x0530, 0x058F, "Armenian"),
    _block(BMP, 0x0590, 0x05FF, "Hebrew"),
    _block(BMP, 0x0600, 0x06FF, "Arabic"),
    _block(BMP, 0x0700, 0x074F, "Syriac"),
    _block(BMP, 0x0750, 0x077F, "Arabic Supplement"),
    _block(BMP, 0x0780, 0x07BF, "Thaana"),
    _block(BMP, 0x07C0, 0x07FF, "NKo"),
    _block(BMP, 0x0800, 0x083F, "Samaritan"),
    _block(BMP, 0x0840, 0x085F, "Mandaic"),
    _block(BMP, 0x0860, 0x086F, "Syriac Supplement"),
    _block(BMP, 0x0870, 0x089F, "Arabic Extended-B"),
    _block(BMP, 0x08A0, 0x08FF, "Arabic Extended-A"),
    _block(BMP, 0x0900, 0x097F, "Devanagari"),
    _block(BMP, 0x0980, 0x09FF, "Bengali"),
    _block(BMP, 0x0A00, 0x0A7F, "Gurmukhi"),
    _block(BMP, 0x0A80, 0x0AFF, "Gujarati"),
    _block(BMP, 0x0B00, 0x0B7F, "Oriya"),
    _block(BMP, 0x0B80, 0x0BFF, "Tamil"),
    _block(BMP, 0x0C00, 0x0C7F, "Telugu"),
    _block(BMP, 0x0C80, 0x0CFF, "Kannada"),
    _block(BMP, 0x0D00, 0x0D7F, "Malayalam"),
    _block(BMP, 0x0D80, 0x0DFF, "Sinhala"),
    _block(BMP, 0x0E00, 0x0E7F, "Thai"),
    _block(BMP, 0x0E80, 0x0EFF, "Lao"),
    _block(BMP, 0x0F00, 0x0FFF, "Tibetan"),
    _block(BMP, 0x1000, 0x109F, "Myanmar"),
    _block(BMP, 0x10A0, 0x10FF, "Georgian"),
    _block(BMP, 0x1100, 0x11FF, "Hangul Jamo"),
    _block(BMP, 0x1200, 0x137F, "Ethiopic"),
    _block(BMP, 0x1380, 0x139F, "Ethiopic Supplement"),
    _block(BMP, 0x13A0, 0x13FF, "Cherokee"),
    _block(BMP, 0x1400, 0x167F, "Unified Canadian Aboriginal Syllabics"),
    _block(BMP, 0x1680, 0x169F, "Ogham"),
    _block(BMP, 0x16A0, 0x16FF, "Runic"),
    _block(BMP, 0x1700, 0x171F, "Tagalog"),
    _block(BMP, 0x1720, 0x173F, "Hanunoo"),
    _block(BMP, 0x1740, 0x175F, "Buhid"),
    _block(BMP, 0x1760, 0x177F, "Tagbanwa"),
    _block(BMP, 0x1780, 0x17FF, "Khmer"),
    _block(BMP, 0x1800, 0x18AF, "Mongolian"),
    _block(BMP, 0x18B0, 0x18FF, "Unified Canadian Aboriginal Syllabics Extended"),
    _block(BMP, 0x1900, 0x194F, "Limbu"),
    _block(BMP, 0x1950, 0x197F, "Tai Le"),
    _block(BMP, 0x1980, 0x19DF, "New Tai Lue"),
    _block(BMP, 0x19E0, 0x19FF, "Khmer Symbols"),
    _block(BMP, 0x1A00, 0x1A1F, "Buginese"),
    _block(BMP, 0x1A20, 0x1AAF, "Tai Tham"),
    _block(BMP, 0x1AB0, 0x1AFF, "Combining Diacritical Marks Extended"),
    _block(BMP, 0x1B00, 0x1B7F, "Balinese"),
    _block(BMP, 0x1B80, 0x1BBF, "Sundanese"),
    _block(BMP, 0x1BC0, 0x1BFF, "Batak"),
    _block(BMP, 0x1C00, 0x1C4F, "Lepcha"),
    _block(BMP, 0x1C50, 0x1C7F, "Ol Chiki"),
    _block(BMP, 0x1C80, 0x1C8F, "Cyrillic Extended-C"),
    _block(BMP, 0x1C90, 0x1CBF, "Georgian Extended"),
    _block(BMP, 0x1CC0, 0x1CCF, "Sundanese Supplement"),
    _block(BMP, 0x1CD0, 0x1CFF, "Vedic Extensions"),
    _block(BMP, 0x1D00, 0x1D7F, "Phonetic Extensions"),
    _block(BMP, 0x1D80, 0x1DBF, "Phonetic Extensions Supplement"),
    _block(BMP, 0x1DC0, 0x1DFF, "Combining Diacritical Marks Supplement"),
    _block(BMP, 0x1E00, 0x1EFF, "Latin Extended Additional"),
    _block(BMP, 0x1F00, 0x1FFF, "Greek Extended"),
    _block(BMP, 0x2000, 0x206F, "General Punctuation"),
    _block(BMP, 0x2070, 0x209F, "Superscripts and Subscripts"),
    _block(BMP, 0x20A0, 0x20CF, "Currency Symbols"),
    _block(BMP, 0x20D0, 0x20FF, "Combining Diacritical Marks for Symbols"),
    _block(BMP, 0x2100, 0x214F, "Letterlike Symbols"),
    _block(BMP, 0x2150, 0x218F, "Number Forms"),
    _block(BMP, 0x2190, 0x21FF, "Arrows"),
    _block(BMP, 0x2200, 0x22FF, "Mathematical Operators"),
    _block(BMP, 0x2300, 0x23FF, "Miscellaneous Technical"),
    _block(BMP, 0x2400, 0x243F, "Control Pictures"),
    _block(BMP, 0x2440, 0x245F, "Optical Character Recognition"),
    _block(BMP, 0x2460, 0x24FF, "Enclosed Alphanumerics"),
    _block(BMP, 0x2500, 0x257F, "Box Drawing"),
    _block(BMP, 0x2580, 0x259F, "Block Elements"),
    _block(BMP, 0x25A0, 0x25FF, "Geometric Shapes"),
    _block(BMP, 0x2600, 0x26FF, "Miscellaneous Symbols"),
    _block(BMP, 0x2700, 0x27BF, "Dingbats"),
    _block(BMP, 0x27C0, 0x27EF, "Miscellaneous Mathematical Symbols-A"),
    _block(BMP, 0x27F0, 0x27FF, "Supplemental Arrows-A"),
    _block(BMP, 0x2800, 0x28FF, "Braille Patterns"),
    _block(BMP, 0x2900, 0x297F, "Supplemental Arrows-B"),
    _block(BMP, 0x2980, 0x29FF, "Miscellaneous Mathematical Symbols-B"),
    _block(BMP, 0x2A00, 0x2AFF, "Supplemental Mathematical Operators"),
    _block(BMP, 0x2B00, 0x2BFF, "Miscellaneous Symbols and Arrows"),
    _block(BMP, 0x2C00, 0x2C5F, "Glagolitic"),
    _block(BMP, 0x2C60, 0x2C7F, "Latin Extended-C"),
    _block(BMP, 0x2C80, 0x2CFF, "Coptic"),
    _block(BMP, 0x2D00, 0x2D2F, "Georgian Supplement"),
    _block(BMP, 0x2D30, 0x2D7F, "Tifinagh"),
    _block(BMP, 0x2D80, 0x2DDF, "Ethiopic Extended"),
    _block(BMP, 0x2DE0, 0x2DFF, "Cyrillic Extended-A"),
    _block(BMP, 0x2E00, 0x2E7F, "Supplemental Punctuation"),
    _block(BMP, 0x2E80, 0x2EFF, "CJK Radicals Supplement"),
    _block(BMP, 0x2F00, 0x2FDF, "Kangxi Radicals"),
    _block(BMP, 0x2FF0, 0x2FFF, "Ideographic Description Characters"),
    _block(BMP, 0x3000, 0x303F, "CJK Symbols and Punctuation"),
    _block(BMP, 0x3040, 0x309F, "Hiragana"),
    _block(BMP, 0x30A0, 0x30FF, "Katakana"),
    _block(BMP, 0x3100, 0x312F, "Bopomofo"),
    _block(BMP, 0x3130, 0x318F, "Hangul Compatibility Jamo"),
    _block(BMP, 0x3190, 0x319F, "Kanbun"),
    _block(BMP, 0x31A0, 0x31BF, "Bopomofo Extended"),
    _block(BMP, 0x31C0, 0x31EF, "CJK Strokes"),
    _block(BMP, 0x31F0, 0x31FF, "Katakana Phonetic Extensions"),
    _block(BMP, 0x3200, 0x32FF, "Enclosed CJK Letters and Months"),
    _block(BMP, 0x3300, 0x33FF, "CJK Compatibility"),
    _block(BMP, 0x3400, 0x4DBF, "CJK Unified Ideographs Extension A"),
    _block(BMP, 0x4DC0, 0x4DFF, "Yijing Hexagram Symbols"),
    _block(BMP, 0x4E00, 0x9FFF, "CJK Unified Ideographs"),
    _block(BMP, 0xA000, 0xA48F, "Yi Syllables"),
    _block(BMP, 0xA490, 0xA4CF, "Yi Radicals"),
    _block(BMP, 0xA4D0, 0xA4FF, "Lisu"),
    _block(BMP, 0xA500, 0xA63F, "Vai"),
    _block(BMP, 0xA640, 0xA69F, "Cyrillic Extended-B"),
    _block(BMP, 0xA6A0, 0xA6FF, "Bamum"),
    _block(BMP, 0xA700, 0xA71F, "Modifier Tone Letters"),
    _block(BMP, 0xA720, 0xA7FF, "Latin Extended-D"),
    _block(BMP, 0xA800, 0xA82F, "Syloti Nagri"),
    _block(BMP, 0xA830, 0xA83F, "Common Indic Number Forms"),
    _block(BMP, 0xA840, 0xA87F, "Phags-pa"),
    _block(BMP, 0xA880, 0xA8DF, "Saurashtra"),
    _block(BMP, 0xA8E0, 0xA8FF, "Devanagari Extended"),
    _block(BMP, 0xA900, 0xA92F, "Kayah Li"),
    _block(BMP, 0xA930, 0xA95F, "Rejang"),
    _block(BMP, 0xA960, 0xA97F, "Hangul Jamo Extended-A"),
    _block(BMP, 0xA980, 0xA9DF, "Javanese"),
    _block(BMP, 0xA9E0, 0xA9FF, "Myanmar Extended-B"),
    _block(BMP, 0xAA00, 0xAA5F, "Cham"),
    _block(BMP, 0xAA60, 0xAA7F, "Myanmar Extended-A"),
    _block(BMP, 0xAA80, 0xAADF, "Tai Viet"),
    _block(BMP, 0xAAE0, 0xAAFF, "Meetei Mayek Extensions"),
    _block(BMP, 0xAB00, 0xAB2F, "Ethiopic Extended-A"),
    _block(BMP, 0xAB30, 0xAB6F, "Latin Extended-E"),
    _block(BMP, 0xAB70, 0xABBF, "Cherokee Supplement"),
    _block(BMP, 0xABC0, 0xABFF, "Meetei Mayek"),
    _block(BMP, 0xAC00, 0xD7AF, "Hangul Syllables"),
    _block(BMP, 0xD7B0, 0xD7FF, "Hangul Jamo Extended-B"),
    _block(BMP, 0xD800, 0xDB7F, "High Surrogates"),
    _block(BMP, 0xDB80, 0xDBFF, "High Private Use Surrogates"),
    _block(BMP, 0xDC00, 0xDFFF, "Low Surrogates"),
    _block(BMP, 0xE000, 0xF8FF, "Private Use Area"),
    _block(BMP, 0xF900, 0xFAFF, "CJK Compatibility Ideographs"),
    _block(BMP, 0xFB00, 0xFB4F, "Alphabetic Presentation Forms"),
    _block(BMP, 0xFB50, 0xFDFF, "Arabic Presentation Forms-A"),
    _block(BMP, 0xFE00, 0xFE0F, "Variation Selectors"),
    _block(BMP, 0xFE10, 0xFE1F, "Vertical Forms"),
    _block(BMP, 0xFE20, 0xFE2F, "Combining Half Marks"),
    _block(BMP, 0xFE30, 0xFE4F, "CJK Compatibility Forms"),
    _block(BMP, 0xFE50, 0xFE6F, "Small Form Variants"),
    _block(BMP, 0xFE70, 0xFEFF, "Arabic Presentation Forms-B"),
    _block(BMP, 0xFF00, 0xFFEF, "Halfwidth and Fullwidth Forms"),
    _block(BMP, 0xFFF0, 0xFFFF, "Specials"),
)

# noinspection SpellCheckingInspection
SMP_BLOCKS: Final[Sequence[UnicodeBlock]] = (
    _block(SMP, 0x10000, 0x1007F, "Linear B Syllabary"),
    _block(SMP, 0x10080, 0x100FF, "Linear B Ideograms"),
    _block(SMP, 0x10100, 0x1013F, "Aegean Numbers"),
    _block(SMP, 0x10140, 0x1018F, "Ancient Greek Numbers"),
    _block(SMP, 0x10190, 0x101CF, "Ancient Symbols"),
    _block(SMP, 0x101D0, 0x101FF, "Phaistos Disc"),
    _block(SMP, 0x10280, 0x1029F, "Lycian"),
    _block(SMP, 0x102A0, 0x102DF, "Carian"),
    _block(SMP, 0x102E0, 0x102FF, "Coptic Epact Numbers"),
    _block(SMP, 0x10300, 0x1032F, "Old Italic"),
    _block(SMP, 0x10330, 0x1034F, "Gothic"),
    _block(SMP, 0x10350, 0x1037F, "Old Permic"),
    _block(SMP, 0x10380, 0x1039F, "Ugaritic"),
    _block(SMP, 0x103A0, 0x103DF, "Old Persian"),
    _block(SMP, 0x10400, 0x1044F, "Deseret"),
    _block(SMP, 0x10450, 0x1047F, "Shavian"),
    _block(SMP, 0x10480, 0x104AF, "Osmanya"),
    _block(SMP, 0x104B0, 0x104FF, "Osage"),
    _block(SMP, 0x10500, 0x1052F, "Elbasan"),
    _block(SMP, 0x10530, 0x1056F, "Caucasian Albanian"),
    _block(SMP, 0x10570, 0x105BF, "Vithkuqi"),
    _block(SMP, 0x105C0, 0x105FF, "Todhri"),
    _block(SMP, 0x10600, 0x1077F, "Linear A"),
    _block(SMP, 0x10780, 0x107BF, "Latin Extended-F"),
    _block(SMP, 0x10800, 0x1083F, "Cypriot Syllabary"),
    _block(SMP, 0x10840, 0x1085F, "Imperial Aramaic"),
    _block(SMP, 0x10860, 0x1087F, "Palmyrene"),
    _block(SMP, 0x10880, 0x108AF, "Nabataean"),
    _block(SMP, 0x108E0, 0x108FF, "Hatran"),
    _block(SMP, 0x10900, 0x1091F, "Phoenician"),
    _block(SMP, 0x10920, 0x1093F, "Lydian"),
    _block(SMP, 0x10980, 0x1099F, "Meroitic Hieroglyphs"),
    _block(SMP, 0x109A0, 0x109FF, "Meroitic Cursive"),
    _block(SMP, 0x10A00, 0x10A5F, "Kharoshthi"),
    _block(SMP, 0x10A60, 0x10A7F, "Old South Arabian"),
    _block(SMP, 0x10A80, 0x10A9F, "Old North Arabian"),
    _block(SMP, 0x10AC0, 0x10AFF, "Manichaean"),
    _block(SMP, 0x10B00, 0x10B3F, "Avestan"),
    _block(SMP, 0x10B40, 0x10B5F, "Inscriptional Parthian"),
    _block(SMP, 0x10B60, 0x10B7F, "Inscriptional Pahlavi"),
    _block(SMP, 0x10B80, 0x10BAF, "Psalter Pahlavi"),
    _block(SMP, 0x10C00, 0x10C4F, "Old Turkic"),
    _block(SMP, 0x10C80, 0x10CFF, "Old Hungarian"),
    _block(SMP, 0x10D00, 0x10D3F, "Hanifi Rohingya"),
    _block(SMP, 0x10D40, 0x10D8F, "Garay"),
    _block(SMP, 0x10E60, 0x10E7F, "Rumi Numeral Symbols"),
    _block(SMP, 0x10E80, 0x10EBF, "Yezidi"),
    _block(SMP, 0x10EC0, 0x10EFF, "Arabic Extended-C"),
    _block(SMP, 0x10F00, 0x10F2F, "Old Sogdian"),
    _block(SMP, 0x10F30, 0x10F6F, "Sogdian"),
    _block(SMP, 0x10F70, 0x10FAF, "Old Uyghur"),
    _block(SMP, 0x10FB0, 0x10FDF, "Chorasmian"),
    _block(SMP, 0x10FE0, 0x10FFF, "Elymaic"),
    _block(SMP, 0x11000, 0x1107F, "Brahmi"),
    _block(SMP, 0x11080, 0x110CF, "Kaithi"),
    _block(SMP, 0x110D0, 0x110FF, "Sora Sompeng"),
    _block(SMP, 0x11100, 0x1114F, "Chakma"),
    _block(SMP, 0x11150, 0x1117F, "Mahajani"),
    _block(SMP, 0x11180, 0x111DF, "Sharada"),
    _block(SMP, 0x111E0, 0x111FF, "Sinhala Archaic Numbers"),
    _block(SMP, 0x11200, 0x1124F, "Khojki"),
    _block(SMP, 0x11280, 0x112AF, "Multani"),
    _block(SMP, 0x112B0, 0x112FF, "Khudawadi"),
    _block(SMP, 0x11300, 0x1137F, "Grantha"),
    _block(SMP, 0x11380, 0x113FF, "Tulu-Tigalari"),
    _block(SMP, 0x11400, 0x1147F, "Newa"),
    _block(SMP, 0x11480, 0x114DF, "Tirhuta"),
    _block(SMP, 0x11580, 0x115FF, "Siddham"),
    _block(SMP, 0x11600, 0x1165F, "Modi"),
    _block(SMP, 0x11660, 0x1167F, "Mongolian Supplement"),
    _block(SMP, 0x11680, 0x116CF, "Takri"),
    _block(SMP, 0x116D0, 0x116FF, "Myanmar Extended-C"),
    _block(SMP, 0x11700, 0x1174F, "Ahom"),
    _block(SMP, 0x11800, 0x1184F, "Dogra"),
    _block(SMP, 0x118A0, 0x118FF, "Warang Citi"),
    _block(SMP, 0x11900, 0x1195F, "Dives Akuru"),
    _block(SMP, 0x119A0, 0x119FF, "Nandinagari"),
    _block(SMP, 0x11A00, 0x11A4F, "Zanabazar Square"),
    _block(SMP, 0x11A50, 0x11AAF, "Soyombo"),
    _block(SMP, 0x11AB0, 0x11ABF, "Unified Canadian Aboriginal Syllabics Extended-A"),
    _block(SMP, 0x11AC0, 0x11AFF, "Pau Cin Hau"),
    _block(SMP, 0x11B00, 0x11B5F, "Devanagari Extended-A"),
    _block(SMP, 0x11BC0, 0x11BFF, "Sunuwar"),
    _block(SMP, 0x11C00, 0x11C6F, "Bhaiksuki"),
    _block(SMP, 0x11C70, 0x11CBF, "Marchen"),
    _block(SMP, 0x11D00, 0x11D5F, "Masaram Gondi"),
    _block(SMP, 0x11D60, 0x11DAF, "Gunjala Gondi"),
    _block(SMP, 0x11EE0, 0x11EFF, "Makasar"),
    _block(SMP, 0x11F00, 0x11F5F, "Kawi"),
    _block(SMP, 0x11FB0, 0x11FBF, "Lisu Supplement"),
    _block(SMP, 0x11FC0, 0x11FFF, "Tamil Supplement"),
    _block(SMP, 0x12000, 0x123FF, "Cuneiform"),
    _block(SMP, 0x12400, 0x1247F, "Cuneiform Numbers and Punctuation"),
    _block(SMP, 0x12480, 0x1254F, "Early Dynastic Cuneiform"),
    _block(SMP, 0x12F90, 0x12FFF, "Cypro-Minoan"),
    _block(SMP, 0x13000, 0x1342F, "Egyptian Hieroglyphs"),
    _block(SMP, 0x13430, 0x1345F, "Egyptian Hieroglyph Format Controls"),
    _block(SMP, 0x13460, 0x143FF, "Egyptian Hieroglyphs Extended-A"),
    _block(SMP, 0x14400, 0x1467F, "Anatolian Hieroglyphs"),
    _block(SMP, 0x16100, 0x1613F, "Gurung Khema"),
    _block(SMP, 0x16800, 0x16A3F, "Bamum Supplement"),
    _block(SMP, 0x16A40, 0x16A6F, "Mro"),
    _block(SMP, 0x16A70, 0x16ACF, "Tangsa"),
    _block(SMP, 0x16AD0, 0x16AFF, "Bassa Vah"),
    _block(SMP, 0x16B00, 0x16B8F, "Pahawh Hmong"),
    _block(SMP, 0x16D40, 0x16D7F, "Kirat Rai"),
    _block(SMP, 0x16E40, 0x16E9F, "Medefaidrin"),
    _block(SMP, 0x16F00, 0x16F9F, "Miao"),
    _block(SMP, 0x16FE0, 0x16FFF, "Ideographic Symbols and Punctuation"),
    _block(SMP, 0x17000, 0x187FF, "Tangut"),
    _block(SMP, 0x18800, 0x18AFF, "Tangut Components"),
    _block(SMP, 0x18B00, 0x18CFF, "Khitan Small Script"),
    _block(SMP, 0x18D00, 0x18D7F, "Tangut Supplement"),
    _block(SMP, 0x1AFF0, 0x1AFFF, "Kana Extended-B"),
    _block(SMP, 0x1B000, 0x1B0FF, "Kana Supplement"),
    _block(SMP, 0x1B100, 0x1B12F, "Kana Extended-A"),
    _block(SMP, 0x1B130, 0x1B16F, "Small Kana Extension"),
    _block(SMP, 0x1B170, 0x1B2FF, "Nushu"),
    _block(SMP, 0x1BC00, 0x1BC9F, "Duployan"),
    _block(SMP, 0x1BCA0, 0x1BCAF, "Shorthand Format Controls"),
    _block(SMP, 0x1CC00, 0x1CEBF, "Symbols for Legacy Computing Supplement"),
    _block(SMP, 0x1CF00, 0x1CFCF, "Znamenny Musical Notation"),
    _block(SMP, 0x1D000, 0x1D0FF, "Byzantine Musical Symbols"),
    _block(SMP, 0x1D100, 0x1D1FF, "Musical Symbols"),
    _block(SMP, 0x1D200, 0x1D24F, "Ancient Greek Musical Notation"),
    _block(SMP, 0x1D2C0, 0x1D2DF, "Kaktovik Numerals"),
    _block(SMP, 0x1D2E0, 0x1D2FF, "Mayan Numerals"),
    _block(SMP, 0x1D300, 0x1D35F, "Tai Xuan Jing Symbols"),
    _block(SMP, 0x1D360, 0x1D37F, "Counting Rod Numerals"),
    _block(SMP, 0x1D400, 0x1D7FF, "Mathematical Alphanumeric Symbols"),
    _block(SMP, 0x1D800, 0x1DAAF, "Sutton SignWriting"),
    _block(SMP, 0x1DF00, 0x1DFFF, "Latin Extended-G"),
    _block(SMP, 0x1E000, 0x1E02F, "Glagolitic Supplement"),
    _block(SMP, 0x1E030, 0x1E08F, "Cyrillic Extended-D"),
    _block(SMP, 0x1E100, 0x1E14F, "Nyiakeng Puachue Hmong"),
    _block(SMP, 0x1E290, 0x1E2BF, "Toto"),
    _block(SMP, 0x1E2C0, 0x1E2FF, "Wancho"),
    _block(SMP, 0x1E4D0, 0x1E4FF, "Nag Mundari"),
    _block(SMP, 0x1E5D0, 0x1E5FF, "Ol Onal"),
    _block(SMP, 0x1E7E0, 0x1E7FF, "Ethiopic Extended-B"),
    _block(SMP, 0x1E800, 0x1E8DF, "Mende Kikakui"),
    _block(SMP, 0x1E900, 0x1E95F, "Adlam"),
    _block(SMP, 0x1EC70, 0x1ECBF, "Indic Siyaq Numbers"),
    _block(SMP, 0x1ED00, 0x1ED4F, "Ottoman Siyaq Numbers"),
    _block(SMP, 0x1EE00, 0x1EEFF, "Arabic Mathematical Alphabetic Symbols"),
    _block(SMP, 0x1F000, 0x1F02F, "Mahjong Tiles"),
    _block(SMP, 0x1F030, 0x1F09F, "Domino Tiles"),
    _block(SMP, 0x1F0A0, 0x1F0FF, "Playing Cards"),
    _block(SMP, 0x1F100, 0x1F1FF, "Enclosed Alphanumeric Supplement"),
    _block(SMP, 0x1F200, 0x1F2FF, "Enclosed Ideographic Supplement"),
    _block(SMP, 0x1F300, 0x1F5FF, "Miscellaneous Symbols and Pictographs"),
    _block(SMP, 0x1F600, 0x1F64F, "Emoticons"),
    _block(SMP, 0x1F650, 0x1F67F, "Ornamental Dingbats"),
    _block(SMP, 0x1F680, 0x1F6FF, "Transport and Map Symbols"),
    _block(SMP, 0x1F700, 0x1F77F, "Alchemical Symbols"),
    _block(SMP, 0x1F780, 0x1F7FF, "Geometric Shapes Extended"),
    _block(SMP, 0x1F800, 0x1F8FF, "Supplemental Arrows-C"),
    _block(SMP, 0x1F900, 0x1F9FF, "Supplemental Symbols and Pictographs"),
    _block(SMP, 0x1FA00, 0x1FA6F, "Chess Symbols"),
    _block(SMP, 0x1FA70, 0x1FAFF, "Symbols and Pictographs Extended-A"),
    _block(SMP, 0x1FB00, 0x1FBFF, "Symbols for Legacy Computing"),
)

SIP_BLOCKS: Final[Sequence[UnicodeBlock]] = (
    _block(SIP, 0x20000, 0x2A6DF, "CJK Unified Ideographs Extension B"),
    _block(SIP, 0x2A700, 0x2B73F, "CJK Unified Ideographs Extension C"),
    _block(SIP, 0x2B740, 0x2B81F, "CJK Unified Ideographs Extension D"),
    _block(SIP, 0x2B820, 0x2CEAF, "CJK Unified Ideographs Extension E"),
    _block(SIP, 0x2CEB0, 0x2EBEF, "CJK Unified Ideographs Extension F"),
    _block(SIP, 0x2EBF0, 0x2EE5F, "CJK Unified Ideographs Extension I"),
    _block(SIP, 0x2F800, 0x2FA1F, "CJK Compatibility Ideographs Supplement"),
)

TIP_BLOCKS: Final[Sequence[UnicodeBlock]] = (
    _block(TIP, 0x30000, 0x3134F, "CJK Unified Ideographs Extension G"),
    _block(TIP, 0x31350, 0x323AF, "CJK Unified Ideographs Extension H"),
)

SSP_BLOCKS: Final[Sequence[UnicodeBlock]] = (
    _block(SSP, 0xE0000, 0xE007F, "Tags"),
    _block(SSP, 0xE0100, 0xE01EF, "Variation Selectors Supplement"),
)

PUA_BLOCKS: Final[Sequence[UnicodeBlock]] = (
    _block(PUA_A, 0xF0000, 0xFFFFF, "Supplementary Private Use Area-A"),
    _block(PUA_B, 0x100000, 0x10FFFF, "Supplementary Private Use Area-B"),
)

BLOCKS: Final[Sequence[UnicodeBlock]] = (
    *BMP_BLOCKS,
    *SMP_BLOCKS,
    *SIP_BLOCKS,
    *TIP_BLOCKS,
    *SSP_BLOCKS,
    *PUA_BLOCKS,
)
